@@ -13,6 +13,7 @@ child processes.
 
 Since the restrictions are set at runtime, from within the process itself,
 you can take into account dynamic information from your configuration.
+
 For example, a server that is supposed to serve files from a specific directory
 can restrict itself to that directory and its subdirectories to mitigate any bugs
 allowing directory traversal attacks. This is much less intrusive than chroot
@@ -30,7 +31,7 @@ about Landlock.
 
       use Linux::Landlock;
 
-      my $ruleset = Linux::Landlock->new(); # this can die
+      my $ruleset = Linux::Landlock->new(supported_abi_version => 4); # this can die
       $ruleset->add_path_rule('/etc/fstab', qw(read_file));
       $ruleset->add_net_rule(22222, qw(bind_tcp));
       $ruleset->apply();
@@ -47,14 +48,27 @@ about Landlock.
 
 =over 1
 
-=item new([handled_fs_actions => \@fs_actions, handled_net_actions => \@net_actions, die_on_unsupported => 1|0])
+=item new([handled_fs_actions => \@fs_actions, handled_net_actions => \@net_actions, restricted_ipc => \@ipc_actions, die_on_unsupported => 1|0])
 
 Create a new L<Linux::Landlock> instance.
 
+C<supported_abi_version> indicates the highest ABI version you want to use. It is highly
+recommended to set this, typically to the version you are testing with.
+The reason is that restrictions added by newer ABI versions might break you program.
+If not set, the running kernel's ABI version is used.
+
 C<handled_fs_actions> and C<handled_net_actions> restrict the set of actions that can be used in rules and that
 will be prevented if not allowed by any rule. By default, all actions supported by the kernel and known to this
 module are covered. This should usually not be changed.
 
+C<restricted_ipc> lists the IPC mechanisms this ruleset should restrict. By default, all IPC mechanisms are
+restricted. Note that this cannot be changed after calling new().
+
+Possible IPC mechanisms are:
+
+    abstract_unix_socket
+    signal
+
 If C<die_on_unsupported> is set to a true value, the module will die if an unsupported access right is requested.
 Otherwise, access rights will be set on a best-effort basis, as intended by the upstream Landlock API design. This
 option should usually not be used.
@@ -94,6 +108,9 @@ Possible access rights are:
 
 See  L<https://docs.kernel.org/userspace-api/landlock.html> for all possible access rights.
 
+Note that B<refer> is special. It is only available starting at ABI version 2, but its restrictions
+are also applied with ABI version 1.
+
 This method dies on error. Errors are: non-existing or non-accessible paths and empty rules.
 If C<die_on_unsupported> is used, it will also die if the rules are not supported by the
 current kernel.
@@ -131,8 +148,8 @@ version 3 (kernel version 6.2 or newer, unless backported).
 
 Network functionality is only available since ABI version 4.
 
-Also keep in mind, that some Perl, or even libc, functions might implicitly rely
-on file system access that could have been restricted by Landlock.
+Also keep in mind, that some Perl modules can implicitly rely on operations
+that are restricted by the Landlock rules you apply, so test carefully.
 
 =head1 AUTHOR
 
@@ -140,7 +157,7 @@ Marc Ballarin, <ballarin.marc@gmx.de>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2024 by Marc Ballarin
+Copyright (C) 2024-2025 by Marc Ballarin
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself.
@@ -149,40 +166,48 @@ it under the same terms as Perl itself.
 
 use strict;
 use warnings;
-use POSIX                   ();
-use List::Util              qw(reduce);
+use POSIX      ();
+use List::Util qw(reduce);
+use Math::BigInt;
 use Linux::Landlock::Direct qw(
   %LANDLOCK_ACCESS_FS
   %LANDLOCK_ACCESS_NET
+  %LANDLOCK_SCOPED
   ll_add_path_beneath_rule
   ll_add_net_port_rule
+  ll_all_fs_access_supported
+  ll_all_net_access_supported
+  ll_all_scoped_supported
   ll_restrict_self
   ll_get_abi_version
-  ll_create_fs_ruleset
-  ll_create_net_ruleset
+  ll_create_ruleset
+  ll_set_max_abi_version
   set_no_new_privs
 );
-our $VERSION = '0.8';
+our $VERSION = '0.9';
 
 sub new {
     my ($class, %args) = @_;
-    die "Landlock is not available\n" if ll_get_abi_version() < 1;
+
+    if ($args{supported_abi_version}) {
+        ll_set_max_abi_version($args{supported_abi_version});
+    }
+    my $kernel_abi_version = ll_get_abi_version();
+    warn "Landlock is not available\n" if $kernel_abi_version < 1;
     my $self = bless {}, $class;
-    $self->{die_on_unsupported}  = $args{die_on_unsupported};
-    $self->{handled_fs_actions}  = ref $args{handled_fs_actions} eq 'ARRAY'  ? $args{handled_fs_actions}  : [];
-    $self->{handled_net_actions} = ref $args{handled_net_actions} eq 'ARRAY' ? $args{handled_net_actions} : [];
+    $self->{die_on_unsupported} = $args{die_on_unsupported};
+    $self->{_rule_fd}           = ll_create_ruleset(
+        __reduce_path_beneath_rules($args{handled_fs_actions}),
+        __reduce_net_port_rules($args{handled_net_actions}),
+        __reduce_ipc_scopes($args{restricted_ipc}),
+    );
     return $self;
 }
 
 sub apply {
     my ($self) = @_;
-    set_no_new_privs() or die "Failed to set no_new_privs: $!\n";
-    if (defined $self->{_fs_fd}) {
-        ll_restrict_self($self->{_fs_fd}) or die "Failed to restrict self: $!\n";
-    }
-    if (defined $self->{_net_fd}) {
-        ll_restrict_self($self->{_net_fd}) or die "Failed to restrict self: $!\n";
-    }
+    set_no_new_privs()                  or die "Failed to set no_new_privs: $!\n";
+    ll_restrict_self($self->{_rule_fd}) or die "Failed to restrict self: $!\n";
     return 1;
 }
 
@@ -193,14 +218,10 @@ sub get_abi_version {
 sub add_path_beneath_rule {
     my ($self, $path, @allowed) = @_;
 
-    unless (defined $self->{_fs_fd}) {
-        $self->{_fs_fd} = ll_create_fs_ruleset(map { $LANDLOCK_ACCESS_FS{ uc $_ } } @{ $self->{handled_fs_actions} })
-          or die "Failed to create ruleset: $!\n";
-    }
     my $is_dir = -d $path;
     if (my $fd = $is_dir ? POSIX::opendir($path) : POSIX::open($path)) {
-        my $allowed = reduce { $a | $b } map { $LANDLOCK_ACCESS_FS{ uc $_ } } @allowed;
-        my $result  = ll_add_path_beneath_rule($self->{_fs_fd}, $allowed, $fd);
+        my $allowed = __reduce_path_beneath_rules(\@allowed);
+        my $result  = ll_add_path_beneath_rule($self->{_rule_fd}, $allowed, $fd);
         if ($is_dir) {
             POSIX::closedir($fd);
         } else {
@@ -216,14 +237,44 @@ sub add_path_beneath_rule {
     }
 }
 
+sub __reduce_path_beneath_rules {
+    if (defined $_[0] && $_[0]->@*) {
+        return reduce { $a | $b } Math::BigInt->bzero,
+          map { $LANDLOCK_ACCESS_FS{ uc $_ } // die "invalid filesystem access right: '$_'\n" } $_[0]->@*;
+    } elsif (defined $_[0]) {
+        return Math::BigInt->bzero;
+    } else {
+        return ll_all_fs_access_supported();
+    }
+}
+
+sub __reduce_net_port_rules {
+    if (defined $_[0] && $_[0]->@*) {
+        return reduce { $a | $b } Math::BigInt->bzero,
+          map { $LANDLOCK_ACCESS_NET{ uc $_ } // die "invalid network access right: '$_'\n" } $_[0]->@*;
+    } elsif (defined $_[0]) {
+        return Math::BigInt->bzero;
+    } else {
+        return ll_all_net_access_supported();
+    }
+}
+
+sub __reduce_ipc_scopes {
+    if (defined $_[0] && $_[0]->@*) {
+        return reduce { $a | $b } Math::BigInt->bzero,
+          map { $LANDLOCK_SCOPED{ uc $_ } // die "invalid IPC mechanism: '$_'\n" } $_[0]->@*;
+    } elsif (defined $_[0]) {
+        return Math::BigInt->bzero;
+    } else {
+        return ll_all_scoped_supported();
+    }
+}
+
 sub add_net_port_rule {
     my ($self, $port, @allowed) = @_;
-    unless (defined $self->{_net_fd}) {
-        $self->{_net_fd} = ll_create_net_ruleset(map { $LANDLOCK_ACCESS_FS{ uc $_ } } @{ $self->{handled_net_actions} })
-          or die "Failed to create ruleset: $!\n";
-    }
-    my $allowed = reduce { $a | $b } map { $LANDLOCK_ACCESS_NET{ uc $_ } } @allowed;
-    my $result  = ll_add_net_port_rule($self->{_net_fd}, $allowed, $port) or die "Failed to add rule: $!\n";
+
+    my $allowed = __reduce_net_port_rules(\@allowed);
+    my $result  = ll_add_net_port_rule($self->{_rule_fd}, $allowed, $port) or die "Failed to add rule: $!\n";
     if ($result != $allowed && $self->{die_on_unsupported}) {
         die "Unsupported access rights: $allowed vs. $result\n";
     }
@@ -252,8 +303,7 @@ sub allow_std_dev_access {
 
 sub DESTROY {
     my ($self) = @_;
-    POSIX::close($self->{_fs_fd})  if defined $self->{_fs_fd};
-    POSIX::close($self->{_net_fd}) if defined $self->{_net_fd};
+    POSIX::close($self->{_rule_fd}) if defined $self->{_rule_fd};
     return;
 }
 
